
#include <mpnt/mpnt.h>
using namespace MpNT;

#include "AES.h"

int AES::get_round_number() {
	if (Nb == 4) {
		if (Nk == 4) {
			return 10;
		} else if (Nk == 6) {
			return 12;
		} else if (Nk == 8) {
			return 14;
		}
	} else if (Nb == 6) {
		if (Nk == 4 || Nk == 6) {
			return 12;
		} else if (Nk == 8) {
			return 14;
		}
	} else if (Nb == 8 && (Nk == 4 || Nk == 6 || Nk == 8)) {
		return 14;
	}
	throw length_error("Wrong Nb/Nk parameters");
}
	
byte AES::get_byte(const MpInt& x, int n) {
	return static_cast<byte>(x.get_wdig(8*n, 8));
}

void AES::set_byte(MpInt& x, int n, byte b) {
	x.set_wdig(8*n, 8, static_cast<Digit>(b));
}

void AES::get_bytes(word w, byte bytes[4]) {
	word mask = 0xFF;
	for (int i = 0; i<4; i++) {
		bytes[i] = static_cast<byte>(w & mask);
		w >>= 8;
	}
}
	
word AES::get_word(const MpInt& x, int n) {
	if (sizeof(Digit) == 32) {
		return x.get_digit(n);
	} else {
		return static_cast<word>(x.get_wdig(32*n, 32));
	}
}

void AES::set_word(MpInt& x, int n, word w) {
	if (sizeof(Digit) == 32) {
		x.set_digit(n, w);
	} else {
		x.set_wdig(32*n, 32, static_cast<Digit>(w));
	}
}
	
void AES::invert_words(const MpInt& x, MpInt& y) {
	Size bits = x.get_length_bin();
	Size n = bits/32 + (bits%32?1:0);
	if (&x == &y) { // in place invert
		for (Size i = 0; i<n/2; i++) {
			word w1 = get_word(x, i);
			word w2 = get_word(x, n-i-1);
			set_word(y, i, w2);
			set_word(y, n-i-1, w1);				
		}
	} else {
		for (Size i = 0; i<n; i++) {
			set_word(y, i, get_word(x, n-i-1));
		}
	}
}

byte AES::mul(byte b1, byte b2) {
	unsigned int product = 0;
	// product = b1 * b2
	int pos = 0;
	while (b2>0) {
		if (1 & b2) {
			product ^= ((unsigned int)b1) << pos;
		}
		pos++;
		b2 >>= 1;
	}
	// product = (b1 * b2) % mx;
	const unsigned int mx = 0x011B; // 9 bits
	unsigned int mask = 0x8000;
	int shift = 7; // 16 - 9
	while (shift >= 0) {
		if (product & mask) {
			product ^= mx << shift;
		}
		mask >>= 1;
		shift--;
	}
	return (byte)product;
}

// generates a random Nk words long key
void AES::generate_key(MpInt& key) {
	key.random_bits(8*4*Nk);
}

//// expands the Nk words key to one that is Nb*(Nr+1) words long
//void AES::expand_key(const MpInt& key, MpInt& w) {
//	for (int i = 0; i<Nk; i++) {
//		set_word(w, i, get_word(key, i/*Nk-i-1*/));
//	}
//	word temp;
//	for (int i = Nk; i< Nb * (Nr+1); i++) {
//		temp = get_word(w, (i + Nb-1-2*((i-1)%Nb))-1/*i-1*/);
//		if (i % Nk == 0) {
//			temp = sub_word(rot_word(temp)) ^ r_con(i/Nk);
//		} else if (Nk > 6 and i % Nk == 4) {
//			temp = sub_word(temp);
//		}
//		set_word(w, i + Nb-1-2*(i%Nb)/*i*/, get_word(w, (i + Nb-1-2*((i-Nk)%Nb))- Nk/*i-Nk*/) ^ temp);
//	}
//}

// expands the Nk words key to one that is Nb*(Nr+1) words long
// Note: each Nb words in w are stored in reversed order
// (relative to the little endian format used for the MpInt)
// they are latter reversed before using the sub_key
void AES::expand_key(const MpInt& key, MpInt& w) {
	for (int i = 0; i<Nk; i++) {
		set_word(w, i, get_word(key, Nk-i-1));
	}
	word temp;
	for (int i = Nk; i< Nb * (Nr+1); i++) {
		temp = get_word(w, i-1);
		if (i % Nk == 0) {
			temp = sub_word(rot_word(temp)) ^ r_con(i/Nk);
		} else if (Nk > 6 and i % Nk == 4) {
			temp = sub_word(temp);
		}
		set_word(w, i, get_word(w, i-Nk) ^ temp);
	}
}

word AES::sub_word(word w) {
	word result;
	byte bytes[4];
	get_bytes(w, bytes);
	for (int i = 3; i>=0; i--) {
		int row = bytes[i] >> 4;
		int column = bytes[i] & 0xF;
		result <<= 8;
		result |= S[row][column];
	}
	return result;
}

word AES::rot_word(word w) {
	return (w << 8) | ((w & 0xFF000000) >> 24);
}

word AES::r_con(int i) {
	return RCi[i] << 24;
}

// x and y are 4*Nb bytes long - the standard says 128 bits (Nb=4)
void AES::encrypt(const MpInt& x, MpInt& y, const MpInt& expanded_key) {
	
	// TODO: test argument lengths - everywhere
	MpInt state(x);
	MpInt key(expanded_key);
//	cout << "expanded_key: " << expanded_key << endl;

	MpInt round_key, mask;
	mask.set_bit(32*Nb, true);
	dec(mask, mask);
//	cout << "mask: " << mask << endl;
	
	// addRoundKey
	bit_and(round_key, key, mask);
	shr(key, key, 32*Nb);
	invert_words(round_key, round_key); // - not needed in the original algorithm
//	cout << "round_key: " << round_key << endl;
	bit_xor(state, state, round_key);
	
	for (int round=1; round<Nr; round++) {
		sub_bytes(state);
		shift_rows(state);
		mix_columns(state);
		
		// addRoundKey
		bit_and(round_key, key, mask);
		shr(key, key, 32*Nb);
		invert_words(round_key, round_key); // - not needed in the original algorithm
//		cout << "round_key: " << round_key << endl;
		bit_xor(state, state, round_key);		
	}
	
	sub_bytes(state);
	shift_rows(state);
	// addRoundKey
	bit_and(round_key, key, mask);
	shr(key, key, 32*Nb);
	invert_words(round_key, round_key); // - not needed in the original algorithm
//	cout << "round_key: " << round_key << endl;
	bit_xor(state, state, round_key);
	
	y = state;
}

void AES::sub_bytes(MpInt& state) {
	for (int i=0; i<4*Nb; i++) {
		byte b = get_byte(state, i);
		int row = b >> 4;
		int column = b & 0xF;
		set_byte(state, i, S[row][column]);
	}
}

void AES::shift_rows(MpInt& state) {
	MpInt result;
	for (int row=0; row<4; row++) {
		for (int j=0; j<Nb; j++) {
			set_byte(result, (4*Nb-1)-(4*j+row),
				get_byte(state, (4*Nb-1)-(4*((j+row)%Nb)+row)));
		}
	}
	state = result;
}

void AES::mix_columns(MpInt& state) {
	MpInt result;
	for (int col = 0; col<Nb; col++) {
		for (int row = 0; row<4; row++) {
			byte b = 0;
			for (int i = 0; i<4; i++) {
				b ^= mul(M2[row][i], get_byte(state, (4*Nb-1)-(4*col+i)));
			}
			set_byte(result, (4*Nb-1)-(4*col+row), b);
		}
	}
	state = result;
}

// x and y are 4*Nb bytes long - the standard says 128 bits (Nb=4)
void AES::decrypt(const MpInt& y, MpInt& x, const MpInt& expanded_key) {
	
	// TODO: test argument lengths - everywhere

	MpInt state(y);
	MpInt key(expanded_key);

	MpInt round_key, mask;
	mask.set_bit(32*Nb, true);
	dec(mask, mask);
	int shift = 32*Nb*Nr;
	shl(mask, mask, shift);
	
	// addRoundKey
	bit_and(round_key, key, mask);
	shr(round_key, round_key, shift);
	shr(mask, mask, 32*Nb);	shift -= 32*Nb;
	invert_words(round_key, round_key); // - not needed in the original algorithm
//	cout << "round_key: " << round_key << endl;
	bit_xor(state, state, round_key);
	
	for (int round=1; round<Nr; round++) {
		shift_rows_inv(state);
		sub_bytes_inv(state);
		
		// addRoundKey
		bit_and(round_key, key, mask);
		shr(round_key, round_key, shift);
		shr(mask, mask, 32*Nb);	shift -= 32*Nb;
		invert_words(round_key, round_key); // - not needed in the original algorithm
//		cout << "round_key: " << round_key << endl;
		bit_xor(state, state, round_key);		

		mix_columns_inv(state);
	}
	
	shift_rows_inv(state);
	sub_bytes_inv(state);

	// addRoundKey
	bit_and(round_key, key, mask);
	shr(round_key, round_key, shift);
	shr(mask, mask, 32*Nb);	shift -= 32*Nb;
	invert_words(round_key, round_key); // - not needed in the original algorithm
//	cout << "round_key: " << round_key << endl;
	bit_xor(state, state, round_key);
	
	x = state;
}

void AES::sub_bytes_inv(MpInt& state) {
	for (int i=0; i<4*Nb; i++) {
		byte b = get_byte(state, i);
		int row = b >> 4;
		int column = b & 0xF;
		set_byte(state, i, S_inv[row][column]);
	}
}

void AES::shift_rows_inv(MpInt& state) {
	MpInt result;
	for (int row=0; row<4; row++) {
		for (int j=0; j<Nb; j++) {
			set_byte(result, (4*Nb-1)-(4*j+row),
				get_byte(state, (4*Nb-1)-(4*((Nb+j-row)%Nb)+row)));
		}
	}
	state = result;
}

void AES::mix_columns_inv(MpInt& state) {
	MpInt result;
	for (int col = 0; col<Nb; col++) {
		for (int row = 0; row<4; row++) {
			byte b = 0;
			for (int i = 0; i<4; i++) {
				b ^= mul(M2_inv[row][i], get_byte(state, (4*Nb-1)-(4*col+i)));
			}
//			shl(result, result, 8);
//			set_byte(result, 0, byte);
			set_byte(result, (4*Nb-1)-(4*col+row), b);
		}
	}
	state = result;
}
