#include <mpnt/mpnt.h>
using namespace MpNT;

#include <vector>
#include <iostream>
#include <string>
#include <exception>
using namespace std;

#include "DESTest.h"

void DESTest::testGenerateKey() {
	CPPUNIT_ASSERT(DES::check_key(key));
}
	
void DESTest::testPC1() {
	MpInt c0d0 = DES::pc1(key);
	CPPUNIT_ASSERT(c0d0.get_length_bin() <= 56);
	for (int i = 0; i<56; i++) {
		CPPUNIT_ASSERT_EQUAL(key.get_bit(DES::PC1[i]-1), c0d0.get_bit(i));
	}
}

void DESTest::testPC2() {
	MpInt cidi;
	cidi.random_bits(56);
	MpInt ki = DES::pc2(cidi);
	CPPUNIT_ASSERT(ki.get_length_bin() <= 48);
	for (int i = 0; i<48; i++) {
		CPPUNIT_ASSERT_EQUAL(cidi.get_bit(DES::PC2[i]-1), ki.get_bit(i));
	}
}

void DESTest::testLSi() {
	for (int i = 1; i<=16; i++) {
		MpInt cd;
		cd.random_bits(56);
//			cout << "cd:     "; cd.write_bin(cout); cout << endl;
		
		MpInt c; // leftmost 28 bits of cd
		MpInt d; // rightmost 28 bits of cd
		shr(c, cd, 28);
		MpInt mask;
		mask.set_bit(28, true);
		dec(mask, mask);
//			cout << "mask:"; mask.write_bin(cout); cout << endl;
		bit_and(d, cd, mask);
		
		bool bit;
		shl(c, c, 1);
		bit = c.get_bit(28);
		c.set_bit(28, false);
		c.set_bit(0, bit);
	
		shl(d, d, 1);
		bit = d.get_bit(28);
		d.set_bit(28, false);
		d.set_bit(0, bit);
		
		if (i == 1 || i == 2 || i == 9 || i == 16) {
			shl(c, c, 1);
			bit = c.get_bit(28);
			c.set_bit(28, false);
			c.set_bit(0, bit);
		
			shl(d, d, 1);
			bit = d.get_bit(28);
			d.set_bit(28, false);
			d.set_bit(0, bit);
		}
		
		MpInt result = c;
		shl(result, result, 28);
		bit_or(result, result, d);

		MpInt result2 = DES::lsi(cd, i);
//			cout << "result :"; result.write_bin(cout); cout << endl;
//			cout << "result2:"; result2.write_bin(cout); cout << endl;
		CPPUNIT_ASSERT_EQUAL(result, result2);
	}
}

void DESTest::testComputeRoundKeys() {
	CPPUNIT_ASSERT_EQUAL(16, (int)round_keys.size());
	for (int i = 1; i<=16; i++) {
		CPPUNIT_ASSERT(round_keys[i-1].get_length_bin() <= 48);
	}
	// TODO: test more?
}

void DESTest::testP0() {
	MpInt x;
	x.random_bits(64);
	MpInt y = DES::p0(x);
	CPPUNIT_ASSERT(y.get_length_bin() <= 64);
	for (int i = 0; i<64; i++) {
		CPPUNIT_ASSERT_EQUAL(x.get_bit(DES::P0[i]-1), y.get_bit(i));
	}
}

void DESTest::testP0Inv() {
	MpInt x;
	x.random_bits(64);
	MpInt y = DES::p0_inv(x);
	CPPUNIT_ASSERT(y.get_length_bin() <= 64);
	for (int i = 0; i<64; i++) {
		CPPUNIT_ASSERT_EQUAL(x.get_bit(i), y.get_bit(DES::P0[i]-1));
	}
}

void DESTest::testP0AndP0Inv() {
	MpInt x;
	x.random_bits(64);
	MpInt y = DES::p0(x);
	MpInt xx = DES::p0_inv(y);
	CPPUNIT_ASSERT_EQUAL(x, xx);
}

void DESTest::testRev() { // test asumes the digit has 32 bits
	MpInt uv;
	uv.random_bits(64);
	MpInt vu;
	vu.set_digit(0, uv.get_digit(1));
	vu.set_digit(1, uv.get_digit(0));
	CPPUNIT_ASSERT_EQUAL(vu, DES::rev(uv));
}

void DESTest::testRev2() {
	MpInt uv;
	uv.random_bits(64);
	MpInt vu = DES::rev(uv);
	MpInt uv_ = DES::rev(vu);
	CPPUNIT_ASSERT_EQUAL(uv, uv_);
}

void DESTest::testE() {
	MpInt v;
	v.random_bits(48);
	MpInt y = DES::e(v);
	CPPUNIT_ASSERT(y.get_length_bin() <= 48);
	for (int i = 0; i<48; i++) {
		CPPUNIT_ASSERT_EQUAL(v.get_bit(DES::E[i]-1), y.get_bit(i));
	}
}

void DESTest::testP() {
	MpInt x;
	x.random_bits(32);
	MpInt y = DES::p(x);
	CPPUNIT_ASSERT(y.get_length_bin() <= 32);
	for (int i = 0; i<32; i++) {
		CPPUNIT_ASSERT_EQUAL(x.get_bit(DES::P[i]-1), y.get_bit(i));
	}
}

void DESTest::testEncryptDecrypt() {
	MpInt x;
	x.random_bits(64);
	MpInt y = DES::encrypt(x, round_keys);
	MpInt xx = DES::decrypt(y, round_keys);
	CPPUNIT_ASSERT_EQUAL(x, xx);
}
