#ifndef MIGNOTTE_H
#define MIGNOTTE_H

#include <mpnt/mpnt.h>
using namespace MpNT;

#include <cmath>
#include <exception>

#include "MignotteSequence.h"
#include "ChineseRemainderTheorem.h"

using namespace std;

class Mignotte 
{
private:
	MignotteSequence &ms;
	
	MpInt S;
	MpModulus q;
	vector<pair<MpInt, int> > I;
public:
	Mignotte(MignotteSequence& ms_) : ms(ms_), I(ms.get_n())
	{
		// TODO: not ok ... should be uniformly distributed between alfa and beta
		unsigned int bits = (ms.get_alpha().get_length_bin()
						+ ms.get_beta().get_length_bin()) / 2;
		do {
			S.random_bits(bits);
		} while (S >= ms.get_alpha() || S <= ms.get_beta());
//		cout << "Secret: " << S << endl;

		for (unsigned int i=0; i<ms.get_n(); i++)
		{
			div_r(I[i].first, S, ms.get_m(i));
			I[i].second = i;
//			cout << "subsecret[" << i << "]=" << I[i].first << endl;
		}
	}

	MignotteSequence get_sequence() {
		return ms;
	}

	unsigned int get_n()
	{	
		return ms.get_n();
	}

	unsigned int get_k()
	{
		return ms.get_k();
	}

	MpInt get_secret()
	{
		return S;
	}

	pair<MpInt, int> get_subsecret(unsigned int i)
	{
		if (i<0 || i>=ms.get_k())
			throw std::out_of_range("index out of range");
		return I[i];
	}

	bool verify( vector<pair<MpInt, int> > users)
	{
		if (users.size() < ms.get_k()) return false;
		MpInt secret;
		vector<MpInt> mods(ms.get_k());
		vector<MpInt> xs(ms.get_k());
		for (unsigned int i=0; i<ms.get_k(); i++)
		{
			mods[i] = ms.get_m( users[i].second );
			xs[i] = users[i].first;
		}
		secret = chinese_remainder_theorem(xs, mods);

//		cout << "Computed secret: " << secret << endl;
//		cout << "Real secret: " << S << endl;
		return (S == secret);
	}
};

#endif
