#include <exception>
using namespace std;

#include <mpnt/mpnt.h>
using namespace MpNT;

#include "Shamir.h"

Shamir::Shamir(unsigned int n_, unsigned int k_, unsigned int bits_)
	: n(n_), k(k_), bits(bits_), I(n_)
{
	MpInt temp;

	gen_prime (temp, bits, 0.00001);
	q.assign(temp);

	S.random_bits(bits);
	red_mod(S, S, q);
//		cout << q.get_value() << endl;
//		cout << S << endl;

	vector<MpInt> coef(k); //a[k-1]...a[1] - random; a0 = S
	coef[0] = S;
	for (unsigned int i=1; i<k; i++)
	{
		coef[i].random_bits(bits);
		red_mod(coef[i], coef[i], q);
//			cout << coef[i] << endl;
	}

	for (unsigned int i=0; i<n; i++)
	{
		bool unique;
		I[i].first = 0;
		do{	
			unique = true;
			(I[i].first).random(bits); 
			red_mod((I[i].first), (I[i].first), q);
			for (unsigned int j=0; j<i; j++)
				if (I[i].first == I[j].first) unique = false;
		}
		while (I[i].first == 0 || !unique);

		I[i].second = 0;
		for (unsigned int j=0; j<k; j++)
		{
			pow_mod(temp, I[i].first, j, q);
			mul_mod(temp, coef[j], temp, q);
			add_mod(I[i].second, I[i].second, temp, q);
		}
//		cout << "subsecret[" << i << "]=" << I[i].first << " " << I[i].second << endl;
	}
}

unsigned int Shamir::get_n()
{	
	return n;
}

unsigned int Shamir::get_k()
{
	return k;
}

unsigned int Shamir::get_bits()
{
	return bits;
}

MpInt Shamir::get_q()
{
	return q.get_value();
}

MpInt Shamir::get_secret()
{
	return S;
}

pair<MpInt, MpInt> Shamir::get_subsecret(unsigned int i)
{
	if (i<0 || i>=n)
		throw std::out_of_range("index out of range");
	return I[i];
}

bool Shamir::verify(vector<pair<MpInt, MpInt> > users)
{
	if (users.size() < k) return false;
	MpInt secret, prod, neg, sum, inv, mul;
	secret = 0;
	for (unsigned int i=0; i<k; i++)
	{
		prod = 1;
		for (unsigned int j=0; j<k; j++)
		{
			if (j != i)
			{
				neg_mod(neg, users[j].first, q);
				add_mod(sum, users[i].first, neg, q);
				inv_mod(inv, sum, q);
				mul_mod(mul, neg, inv, q);
				mul_mod(prod, prod, mul, q);
			}
		}
		mul_mod(prod, users[i].second, prod, q);
		add_mod(secret, secret, prod, q);
	}
//		cout << "Computed secret: " << secret << endl;
//		cout << "Real secret:     " << S << endl;
	return (S == secret);
}
