#include <mpnt/mpnt.h>
using namespace MpNT;

#include <iostream>
using namespace std;

#include "ShamirTest.h"

void ShamirTest::setUp() {
	n = 10;
	k = 5;
	bits = 100;
	shamir = new Shamir(n, k, bits);
}
		
void ShamirTest::tearDown() {
	delete shamir;
}

void ShamirTest::testConstructorBasic() {
	CPPUNIT_ASSERT_EQUAL(n, shamir->get_n());
	CPPUNIT_ASSERT_EQUAL(k, shamir->get_k());
	CPPUNIT_ASSERT_EQUAL(bits, shamir->get_bits());
}

void ShamirTest::testQPrime() {
	CPPUNIT_ASSERT(miller_rabin(shamir->get_q(), 1000));
}

void ShamirTest::testSecret() {
	MpInt secret = shamir->get_secret();
	CPPUNIT_ASSERT(cmp(secret, 0) >= 0);
	CPPUNIT_ASSERT(cmp(secret, shamir->get_q()) < 0);
}

void ShamirTest::testSubsecretsBasic() {
	vector<pair<MpInt, MpInt> > subsecrets;
	subsecrets.reserve(shamir->get_n());
	for (unsigned int i=0; i<shamir->get_n(); i++) {
		pair<MpInt, MpInt> sub_secret = shamir->get_subsecret(i);
//			cout << "sub_secret[" << i << "]=(" << sub_secret.first
//										<< "," << sub_secret.second << endl;
		CPPUNIT_ASSERT(cmp(sub_secret.first, 0) > 0);
		CPPUNIT_ASSERT(cmp(sub_secret.first, shamir->get_q()) < 0);
		for (unsigned int j=0; j<i; j++) {
			CPPUNIT_ASSERT(subsecrets.at(j).first != sub_secret.first);
		}
		subsecrets.push_back(sub_secret);
	}
}

void ShamirTest::testGetSubsecretException() throw (std::out_of_range)  {
	shamir->get_subsecret(shamir->get_n());
}

void ShamirTest::testSubsecretsValid() {
	vector<pair<MpInt, MpInt> > users(k);
	for (unsigned int i=0; i<k; i++) {
		users[i] = shamir->get_subsecret(i);
	}
	CPPUNIT_ASSERT(shamir->verify(users));
}

void ShamirTest::testSubsecretsValid2() {
	vector<pair<MpInt, MpInt> > users;
	users.reserve(shamir->get_n());
	for (unsigned int i=n-k; i<n; i++) {
		users.push_back(shamir->get_subsecret(i));
	}
	CPPUNIT_ASSERT(shamir->verify(users));
}

void ShamirTest::testSubsecretsInvalid() {
	vector<pair<MpInt, MpInt> > users(k);
	for (unsigned int i=0; i<k; i++) {
		users[i] = shamir->get_subsecret(i);
	}
	add(users[0].second, users[0].second, 100);
	CPPUNIT_ASSERT(shamir->verify(users));
}	

void ShamirTest::testSubsecretsInvalid2() {
	vector<pair<MpInt, MpInt> > users(k);
	for (unsigned int i=0; i<k-1; i++) {
		users[i] = shamir->get_subsecret(i);
	}
	CPPUNIT_ASSERT(shamir->verify(users));
}	
