
#include <vector>
#include <string>
#include <cstdlib>
#include <exception>
#include <iostream>
#include <cmath>
using namespace std;

#include "Vigenere.h"
	
const float Vigenere::english_freq[] =
	{  /* a */
	  7.81, 1.28, 2.93, 4.11, 13.05, 2.88, 1.39,
	  5.85, 6.77, 0.23, 0.42, 3.60, 2.62, 7.28,
	  8.21, 2.15, 0.14, 6.64, 6.46, 9.02, 2.77,
	  1.00, 1.49, 0.30, 1.51, 0.09
	}; /* z */
	
void Vigenere::convert(string str, vector<int>& arr) {
	arr.clear();
	for (unsigned int i=0; i<str.size(); i++) {
		if (str[i] < 'a' || str[i] > 'z') {
			throw "character not in [a..z]";
//			throw out_of_range("character not in [a..z]");
		} else {
			arr.push_back((int)(str[i]-'a'));
		}
	}
}

void Vigenere::convert(vector<int> arr, string& str) {
	str.clear();
	for (unsigned int i=0; i<arr.size(); i++) {
		str.push_back((char)('a'+arr[i]));
	}
}


void Vigenere::gen_key(int m, vector<int>& arr) {
	arr.clear();	
	for (int i=0; i<m; i++) {
		arr.push_back(rand() % 26);
	}
}

void Vigenere::clear_text(const string& src, string& dest) {
	dest.clear();	
	for (unsigned int i=0; i<src.size(); i++) {
		if (src[i] >= 'a' && src[i] <= 'z') {
			dest.push_back(src[i]);
		}
	}
}

void Vigenere::m_by_m_starting_j(const string& src, int j, int m, string& dest) {
	dest.clear();	
	for (unsigned int i=j; i<src.size(); i += m) {
		dest.push_back(src[i]);
	}	
}

double Vigenere::IC(string s)
{
	double ic = 0;
	double no;
	for (int i=0; i<26; i++)
	{
		no = 0;
		for (unsigned int j=0; j<s.size(); j++)
			if (s[j] == ((char)('a'+i))) 
				no += 1;
		ic += (no/s.size())*((no-1)/(s.size()-1));
	}
	return ic;
}

//int Vigenere::compute_key_length(const string& y)
//{
//	unsigned int m = 1;
//	bool key_lenght_found;
//	string s;
//	double ic;
//	do{
//		m++;
//		key_lenght_found = true;
//		for (unsigned int i=0; i<m && key_lenght_found; i++)
//		{
//			m_by_m_starting_j(y, i, m, s);
//			ic = IC(s);
////			cout << "m=" << m << " i=" << i << " ic=" << ic << endl;
//			if (abs(ic-IC_ENGLISH) > DELTA)
//				key_lenght_found = false;
//		}
//	} 
//	while (!key_lenght_found && m < 100);
//	return m;
//}

int Vigenere::compute_key_length(const string& y)
{
	unsigned int m = 0;
	bool key_lenght_found;
	vector<double> ic;
	string s;
	double avg, disp;
	do{
		ic.clear();
		m++;
		key_lenght_found = true;
		avg = 0;
		for (unsigned int i=0; i<m; i++)
		{
			m_by_m_starting_j(y, i, m, s);
			ic.push_back( IC(s) );
			avg += ic[i];
		}
		avg = avg / m;
		disp = 0;
		for (unsigned int i=0; i<m; i++)
			disp += (avg - ic[i])*(avg - ic[i]);
		disp = disp / m;
		//cout << "m=" << m << " avg: " << avg << " disp: " << disp << endl;
		if ( abs(avg-IC_ENGLISH) > DELTA_AVG || disp > DELTA_DISP)
			key_lenght_found = false;
	} 
	while (!key_lenght_found && m < 100); // TODO: find a better limit here
	return m;
}

void Vigenere::shift(const string& src, int k, string& dest)
{
	dest.clear();
	for (unsigned int i=0; i<src.size(); i++)
	{
		dest.push_back((char)(((src[i]-'a')+k)%26 + 'a')); 
	}	
}

double Vigenere::MIC(string s)
{
	//english_freq/100
	double mic = 0;
	double no;
	for (int i=0; i<26; i++)
	{
		no = 0;
		for (unsigned int j=0; j<s.size(); j++)
			if (s[j] == ((char)('a'+i))) 
				no += 1;
		mic += (english_freq[i]/100)*(no/s.size());
	}
	return mic;
}

void Vigenere::atack(const string& source, vector<int>& key)
{
	key.clear();
	int m = compute_key_length(source);
	string ymj, shifted_ymj;
	double mic;
	for (int i=0; i<m; i++)
	{
		double max_MIC = 0.0;
		int index_max_MIC = -1;
		for (int s=0; s<26; s++)
		{
			m_by_m_starting_j(source, i, m, ymj);
			shift(ymj, s, shifted_ymj);
			mic = MIC(shifted_ymj);
			if (mic > max_MIC)
			{
				max_MIC = mic;
				index_max_MIC = s;
			}
		}
		key.push_back( (26-index_max_MIC)%26 );
	}
}

