#ifndef _VIGENERE_H_
#define _VIGENERE_H_

#include <vector>
#include <string>
using namespace std;

class Vigenere {
private:
	int key_size;
	vector<int> key;
	
public:
	static const double DELTA = 0.01;
	static const double DELTA_AVG = 0.001;
	static const double DELTA_DISP = 0.0001;
	static const double IC_ENGLISH = 0.065;
	
	static const float english_freq[];

	Vigenere(int m): key_size(m), key(m) {
		gen_key(key_size, key);
	}
	
	int get_key_size() {
		return key_size;
	}
	
	vector<int> get_key() {
		return key;
	}
	
	void encrypt(vector<int> source, vector<int>& dest) {
		dest.clear();
		for (unsigned int i=0; i<source.size(); i++) {
			dest.push_back((source[i] + key[i%key_size]) % 26);
		}
	}

	void decrypt(vector<int> source, vector<int>& dest) {
		dest.clear();
		for (unsigned int i=0; i<source.size(); i++) {
			dest.push_back((source[i] - key[i%key_size] + 26) % 26);
		}
	}
	
	void encrypt(string source, string& dest) {
		vector<int> x, y;
		convert(source, x);
		encrypt(x, y);
		convert(y, dest);
	}	

	void decrypt(string source, string& dest) {
		vector<int> x, y;
		convert(source, x);
		decrypt(x, y);
		convert(y, dest);
	}
	
	// atack on Vigenere 
	static int compute_key_length(const string& y);

	static void convert(string str, vector<int>& arr);
	static void convert(vector<int> arr, string& str);
	static void gen_key(int m, vector<int>& key);
	static void clear_text(const string& src, string& dest);
	static double IC(string s);
	static void m_by_m_starting_j(const string& src, int j, int m, string& dest);
	static void shift(const string& src, int k, string& dest);
	static double MIC(string s);
	static void atack(const string& source, vector<int>& key);
};

#endif //_VIGENERE_H_
